@@ -1,30 +1,34 @@
+import datetime
+import uuid
+from functools import partial
+from typing import Any, Dict, Union
+
+import numpy as np
+import pandas as pd
+from fs.base import FS
+from fs.errors import ResourceNotFound
+from fs.memoryfs import MemoryFS
+
 from .constants import DEFAULT_LICENSES
 from .errors import (
     Closed,
+    InvalidMimetype,
     LengthMismatch,
     NonUnique,
-    InvalidMimetype,
     PotentialInconsistency,
+    ShapeMismatch,
+    WrongDatatype,
 )
 from .filesystem import clean_datapackage_name
-from .io_helpers import file_writer, file_reader
-from .proxies import UndefinedInterface, Proxy
+from .io_helpers import file_reader, file_writer
+from .proxies import Proxy, UndefinedInterface
 from .utils import (
+    NoneSorter,
     check_name,
     check_suffix,
     load_bytes,
     resolve_dict_iterator,
-    NoneSorter,
 )
-from fs.base import FS
-from fs.errors import ResourceNotFound
-from fs.memoryfs import MemoryFS
-from functools import partial
-from typing import Union, Any
-import datetime
-import numpy as np
-import pandas as pd
-import uuid
 
 
 class DatapackageBase:
@@ -111,7 +115,9 @@ class DatapackageBase:
         del self.data[index]
 
     def del_resource_group(self, name: str) -> None:
-        """Remove a resource group, and delete its data files, if any."""
+        """Remove a resource group, and delete its data files, if any.
+
+        Use ``exclude_resource_group`` if you want to keep the underlying resource in the filesystem."""
         if self._modified:
             raise PotentialInconsistency(
                 "Datapackage is modified; save modifications or reload"
@@ -182,10 +188,18 @@ class DatapackageBase:
             fdp.indexer = self.indexer
         return fdp
 
-    def exclude_resource_group(self, group: str) -> "FilteredDatapackage":
-        """Filter a datapackage to exclude a resource group by group name.
+    def exclude(self, filters: Dict[str, str]) -> "FilteredDatapackage":
+        """Filter a datapackage to exclude resources matching a filter.
 
-        Use ``filter_by_attribute("group", "group_name")`` to get only this resource group.
+        Usage cases:
+
+        Filter out a given resource:
+
+            exclude_generic({"matrix': "some_label"})
+
+        Filter out a resource group with a given kind:
+
+            exclude_generic({"group': "some_group", "kind": "some_kind"})
 
         """
         fdp = FilteredDatapackage()
@@ -195,39 +209,15 @@ class DatapackageBase:
         if hasattr(self, "indexer"):
             fdp.indexer = self.indexer
 
-        to_include = [
+        indices_to_include = [
             i
             for i, resource in enumerate(self.resources)
-            if resource.get("group") != group
+            if any(resource.get(key) != value for key, value in filters.items())
         ]
-        fdp.data = [o for i, o in enumerate(self.data) if i in to_include]
-        fdp.resources = [o for i, o in enumerate(self.resources) if i in to_include]
-        return fdp
-
-    def exclude_resource_group_kind(
-        self, group: str, kind: str
-    ) -> "FilteredDatapackage":
-        """Filter a datapackage to exclude a certain kinds of resources in a given resource group.
-
-        The intended use case is to do complicated calculation setups, like distributions from one resource group but just the static array from a different resource group in the same datapackage.
-
-        ``group`` and ``kind`` are the group and kind labels of the resource to exclude.
-
-        """
-        fdp = FilteredDatapackage()
-        fdp.metadata = {k: v for k, v in self.metadata.items() if k != "resources"}
-        fdp.metadata["resources"] = []
-
-        if hasattr(self, "indexer"):
-            fdp.indexer = self.indexer
-
-        to_exclude = [
-            i
-            for i, resource in enumerate(self.resources)
-            if resource.get("group") == group and resource.get("kind") == kind
+        fdp.data = [o for i, o in enumerate(self.data) if i in indices_to_include]
+        fdp.resources = [
+            o for i, o in enumerate(self.resources) if i in indices_to_include
         ]
-        fdp.data = [o for i, o in enumerate(self.data) if i not in to_exclude]
-        fdp.resources = [o for i, o in enumerate(self.resources) if i not in to_exclude]
         return fdp
 
 
@@ -442,8 +432,9 @@ class Datapackage(DatapackageBase):
         )
         name = self._prepare_name(name)
 
+        indices_array = load_bytes(indices_array)
         self._add_numpy_array_resource(
-            array=load_bytes(indices_array),
+            array=indices_array,
             name=name + ".indices",
             group=name,
             kind="indices",
@@ -451,8 +442,21 @@ class Datapackage(DatapackageBase):
             **kwargs,
         )
         if data_array is not None:
+            data_array = load_bytes(data_array)
+            if len(data_array.shape) > 1:
+                raise ShapeMismatch(
+                    "Passed {}-d array to 1-d function `add_persistent_vector`".format(
+                        len(data_array.shape)
+                    )
+                )
+            elif data_array.shape != indices_array.shape:
+                raise ShapeMismatch(
+                    "`data_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                        data_array.shape, indices_array.shape
+                    )
+                )
             self._add_numpy_array_resource(
-                array=load_bytes(data_array),
+                array=data_array,
                 group=name,
                 name=name + ".data",
                 kind="data",
@@ -460,27 +464,49 @@ class Datapackage(DatapackageBase):
                 **kwargs,
             )
         if distributions_array is not None:
+            distributions_array = load_bytes(distributions_array)
             # If no uncertainty, don't need to store it
             if (distributions_array["uncertainty_type"] < 2).sum() < len(
                 distributions_array
             ):
+                if distributions_array.shape != indices_array.shape:
+                    raise ShapeMismatch(
+                        "`distributions_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                            distributions_array.shape, indices_array.shape
+                        )
+                    )
                 self._add_numpy_array_resource(
-                    array=load_bytes(distributions_array),
+                    array=distributions_array,
                     name=name + ".distributions",
                     group=name,
                     kind="distributions",
                     keep_proxy=keep_proxy,
                     **kwargs,
                 )
-        if flip_array is not None and flip_array.sum():
-            self._add_numpy_array_resource(
-                array=load_bytes(flip_array),
-                group=name,
-                name=name + ".flip",
-                kind="flip",
-                keep_proxy=keep_proxy,
-                **kwargs,
-            )
+        if flip_array is not None:
+            flip_array = load_bytes(flip_array)
+            # If no flips, don't need to store it
+            if flip_array.sum():
+                if flip_array.dtype != bool:
+                    raise WrongDatatype(
+                        "`flip_array` dtype is {}, but must be `bool`".format(
+                            flip_array.dtype
+                        )
+                    )
+                elif flip_array.shape != indices_array.shape:
+                    raise ShapeMismatch(
+                        "`flip_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                            flip_array.shape, indices_array.shape
+                        )
+                    )
+                self._add_numpy_array_resource(
+                    array=flip_array,
+                    group=name,
+                    name=name + ".flip",
+                    kind="flip",
+                    keep_proxy=keep_proxy,
+                    **kwargs,
+                )
 
     def add_persistent_array(
         self,
@@ -501,31 +527,60 @@ class Datapackage(DatapackageBase):
         )
         name = self._prepare_name(name)
 
+        indices_array = load_bytes(indices_array)
         self._add_numpy_array_resource(
-            array=load_bytes(data_array),
-            name=name + ".data",
-            group=name,
-            kind="data",
-            keep_proxy=keep_proxy,
-            **kwargs,
-        )
-        self._add_numpy_array_resource(
-            array=load_bytes(indices_array),
+            array=indices_array,
             name=name + ".indices",
             kind="indices",
             group=name,
             keep_proxy=keep_proxy,
             **kwargs,
         )
-        if flip_array is not None and flip_array.sum():
-            self._add_numpy_array_resource(
-                array=load_bytes(flip_array),
-                group=name,
-                name=name + ".flip",
-                kind="flip",
-                keep_proxy=keep_proxy,
-                **kwargs,
+
+        data_array = load_bytes(data_array)
+        if len(data_array.shape) != 2:
+            raise ShapeMismatch(
+                "Passed {}-d array to 2-d function `add_persistent_array`".format(
+                    len(data_array.shape)
+                )
             )
+        elif data_array.shape[0] != indices_array.shape[0]:
+            raise ShapeMismatch(
+                "`data_array` row number ({}) doesn't match `indices_array` ({}).".format(
+                    data_array.shape, indices_array.shape
+                )
+            )
+        self._add_numpy_array_resource(
+            array=data_array,
+            name=name + ".data",
+            group=name,
+            kind="data",
+            keep_proxy=keep_proxy,
+            **kwargs,
+        )
+        if flip_array is not None:
+            flip_array = load_bytes(flip_array)
+            if flip_array.sum():
+                if flip_array.dtype != bool:
+                    raise WrongDatatype(
+                        "`flip_array` dtype is {}, but must be `bool`".format(
+                            flip_array.dtype
+                        )
+                    )
+                elif flip_array.shape != indices_array.shape:
+                    raise ShapeMismatch(
+                        "`flip_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                            flip_array.shape, indices_array.shape
+                        )
+                    )
+                self._add_numpy_array_resource(
+                    array=flip_array,
+                    group=name,
+                    name=name + ".flip",
+                    kind="flip",
+                    keep_proxy=keep_proxy,
+                    **kwargs,
+                )
 
     def write_modified(self):
         """Write the data in modified files to the filesystem (if allowed)."""
@@ -604,25 +659,38 @@ class Datapackage(DatapackageBase):
         )
         name = self._prepare_name(name)
 
-        # Do something with dynamic vector
-
+        indices_array = load_bytes(indices_array)
         self._add_numpy_array_resource(
-            array=load_bytes(indices_array),
+            array=indices_array,
             name=name + ".indices",
             group=name,
             kind="indices",
             keep_proxy=keep_proxy,
             **kwargs,
         )
-        if flip_array is not None and flip_array.sum():
-            self._add_numpy_array_resource(
-                array=load_bytes(flip_array),
-                group=name,
-                name=name + ".flip",
-                kind="flip",
-                keep_proxy=keep_proxy,
-                **kwargs,
-            )
+        if flip_array is not None:
+            flip_array = load_bytes(flip_array)
+            if flip_array.sum():
+                if flip_array.dtype != bool:
+                    raise WrongDatatype(
+                        "`flip_array` dtype is {}, but must be `bool`".format(
+                            flip_array.dtype
+                        )
+                    )
+                elif flip_array.shape != indices_array.shape:
+                    raise ShapeMismatch(
+                        "`flip_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                            flip_array.shape, indices_array.shape
+                        )
+                    )
+                self._add_numpy_array_resource(
+                    array=flip_array,
+                    group=name,
+                    name=name + ".flip",
+                    kind="flip",
+                    keep_proxy=keep_proxy,
+                    **kwargs,
+                )
 
         self.data.append(interface)
         resource = {
@@ -656,8 +724,9 @@ class Datapackage(DatapackageBase):
         )
         name = self._prepare_name(name)
 
+        indices_array = load_bytes(indices_array)
         self._add_numpy_array_resource(
-            array=load_bytes(indices_array),
+            array=indices_array,
             name=name + ".indices",
             group=name,
             kind="indices",
@@ -665,14 +734,28 @@ class Datapackage(DatapackageBase):
             **kwargs,
         )
         if flip_array is not None:
-            self._add_numpy_array_resource(
-                array=load_bytes(flip_array),
-                group=name,
-                name=name + ".flip",
-                kind="flip",
-                keep_proxy=keep_proxy,
-                **kwargs,
-            )
+            flip_array = load_bytes(flip_array)
+            if flip_array.sum():
+                if flip_array.dtype != bool:
+                    raise WrongDatatype(
+                        "`flip_array` dtype is {}, but must be `bool`".format(
+                            flip_array.dtype
+                        )
+                    )
+                elif flip_array.shape != indices_array.shape:
+                    raise ShapeMismatch(
+                        "`flip_array` shape ({}) doesn't match `indices_array` ({}).".format(
+                            flip_array.shape, indices_array.shape
+                        )
+                    )
+                self._add_numpy_array_resource(
+                    array=flip_array,
+                    group=name,
+                    name=name + ".flip",
+                    kind="flip",
+                    keep_proxy=keep_proxy,
+                    **kwargs,
+                )
 
         self.data.append(interface)
         resource = {
